/*	$Id: ueagle_mod.c,v 1.1 2004/08/29 19:07:56 damien Exp $	*/

/*-
 * Copyright (c) 2003, 2004
 *	Damien Bergamini <damien.bergamini@free.fr>. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if defined(__FreeBSD__) || defined(__NetBSD__)
#include <sys/cdefs.h>
#if defined(__NetBSD__)
__KERNEL_RCSID(0, "$Id: ueagle_mod.c,v 1.1 2004/08/29 19:07:56 damien Exp $");
#elif defined(__FreeBSD__)
__FBSDID("$Id: ueagle_mod.c,v 1.1 2004/08/29 19:07:56 damien Exp $");
#endif
#endif

/*-
 * Analog Devices Eagle chipset driver
 * http://www.analog.com/
 */

#if defined(__NetBSD__)
#include "opt_inet.h"
#include "opt_natm.h"
#include "bpfilter.h"
#elif defined(__OpenBSD__)
#include "bpfilter.h"
#elif defined(__FreeBSD__)
#include "opt_inet.h"
#include "opt_natm.h"
#endif

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#if defined(__NetBSD__) || defined(__OpenBSD__)
#include <sys/device.h>
#elif defined(__FreeBSD__)
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/ioccom.h>
#include <sys/uio.h>
#endif
#include <sys/conf.h>
#include <sys/file.h>
#include <sys/fcntl.h>
#include <sys/mbuf.h>
#include <sys/sockio.h>
#include <sys/socket.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_atm.h>
#if defined(__FreeBSD__) && __FreeBSD_version >= 501000
#include <net/if_media.h>
#endif

#if defined(__FreeBSD__) || NBPFILTER > 0
#include <net/bpf.h>
#endif

#ifdef INET
#include <netinet/in.h>
#include <netinet/if_atm.h>
#endif

#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#include <dev/usb/usbcdc.h>

#include "ueaglereg.h"
#include "ueaglevar.h"
#include "ueagle_fw.h"

#ifdef UEAGLE_DEBUG
#define DPRINTF(x)	if (ueagledebug > 0) logprintf x
#define DPRINTFN(n, x)	if (ueagledebug > (n)) logprintf x
int ueagledebug = 0;
#if defined(__FreeBSD__)
SYSCTL_NODE(_hw_usb, OID_AUTO, ueagle, CTLFLAG_RW, 0, "USB ueagle");
SYSCTL_INT(_hw_usb_ueagle, OID_AUTO, debug, CTLFLAG_RW, &ueagledebug, 0,
    "ueagle debug level");
#endif
#else
#define DPRINTF(x)
#define DPRINTFN(n, x)
#endif

/* various supported device vendors/products */
Static const struct usb_devno ueagle_devs[] = {
	/* EagleI/II/IIC */
	{ USB_VENDOR_ADI, USB_PRODUCT_ADI_EAGLEI },
	{ USB_VENDOR_ADI, USB_PRODUCT_ADI_EAGLEII },
	{ USB_VENDOR_ADI, USB_PRODUCT_ADI_EAGLEIIC },

	/* U.S. Robotics */
	{ USB_VENDOR_USR, USB_PRODUCT_USR_MILLER_A },
	{ USB_VENDOR_USR, USB_PRODUCT_USR_MILLER_B },
	{ USB_VENDOR_USR, USB_PRODUCT_USR_HEINEKEN_A },
	{ USB_VENDOR_USR, USB_PRODUCT_USR_HEINEKEN_B }
};

Static int ueagle_getesi(struct ueagle_softc *, u_char *);
Static void ueagle_idma_cb(usbd_xfer_handle, usbd_private_handle, usbd_status);
#if defined(__NetBSD__) || defined(__OpenBSD__)
Static void ueagle_loadpage(void *);
#elif defined(__FreeBSD__)
Static void ueagle_loadpage(struct ueagle_softc *, u_int16_t, u_int16_t);
#endif
Static void ueagle_dispatch_cmv(struct ueagle_softc *, cmv_t *);
Static usbd_status ueagle_request(struct ueagle_softc *, u_int16_t, u_int16_t,
    u_int16_t, void *);
Static usbd_status ueagle_cmv(struct ueagle_softc *, u_int8_t, u_int32_t,
    u_int16_t, u_int32_t, cmv_cb_f);
Static void ueagle_stat_cb(struct ueagle_softc *, int, u_int32_t);
Static void ueagle_stat(void *);
Static void ueagle_cntl02_cb(struct ueagle_softc *, int, u_int32_t);
Static void ueagle_sync(void *);
Static void ueagle_optn_cb(struct ueagle_softc *, int, u_int32_t);
Static void ueagle_bootstrap(void *);
Static void ueagle_endreset(void *);
Static void ueagle_startreset(struct ueagle_softc *);
Static void ueagle_boot(struct ueagle_softc *);
Static void ueagle_intr(usbd_xfer_handle, usbd_private_handle, usbd_status);
Static void ueagle_crc_init(void);
Static u_int32_t ueagle_crc_update(u_int32_t, u_int8_t *, int);
Static struct ueagle_vcc *ueagle_find_vcc(struct ueagle_softc *, u_int8_t,
    u_int16_t);
Static void ueagle_record(struct ueagle_softc *, uint8_t *);
Static void ueagle_rx_cb(usbd_xfer_handle, usbd_private_handle, usbd_status);
Static void ueagle_tx_cb(usbd_xfer_handle, usbd_private_handle, usbd_status);
Static int ueagle_send(struct ueagle_softc *, struct mbuf *, int);
Static void ueagle_start(struct ifnet *);
Static int ueagle_tx_init(struct ueagle_softc *);
Static void ueagle_tx_stop(struct ueagle_softc *);
Static int ueagle_rx_init(struct ueagle_softc *);
Static void ueagle_rx_stop(struct ueagle_softc *);
Static void ueagle_init(void *);
Static void ueagle_stop(struct ueagle_softc *);
Static int ueagle_open_vcc(struct ueagle_softc *, caddr_t);
Static int ueagle_close_vcc(struct ueagle_softc *, caddr_t);
Static int ueagle_ioctl(struct ifnet *, u_long, caddr_t);

USB_DECLARE_DRIVER(ueagle);

USB_MATCH(ueagle)
{
	USB_MATCH_START(ueagle, uaa);

	if (uaa->iface != NULL)
		return UMATCH_NONE;

	return (usb_lookup(ueagle_devs, uaa->vendor, uaa->product) != NULL) ?
	    UMATCH_VENDOR_PRODUCT : UMATCH_NONE;
}

USB_ATTACH(ueagle)
{
	USB_ATTACH_START(ueagle, sc, uaa);
	char devinfo[1024];
	u_char esi[6];
	struct ifnet *ifp;
	char *errmsg;

#if defined(__FreeBSD__)
	bzero(sc, sizeof (struct ueagle_softc));
#endif
	sc->sc_udev = uaa->device;

#if defined(__FreeBSD__)
	usbd_devinfo(sc->sc_udev, 0, devinfo);
#elif defined(__NetBSD__) || defined(__OpenBSD__)
	usbd_devinfo(sc->sc_udev, 0, devinfo, sizeof devinfo);
#endif
	USB_ATTACH_SETUP;
#if defined(__NetBSD__) || defined(__OpenBSD__)
	printf("%s: %s\n", USBDEVNAME(sc->sc_dev), devinfo);
#endif

	usb_callout_init(sc->sc_ch);
#if defined(__NetBSD__) || defined(__OpenBSD__)
	usb_init_task(&sc->sc_task, ueagle_loadpage, sc);
#endif

	if (usbd_set_config_no(sc->sc_udev, UEAGLE_CONFIG_NO, 0) != 0) {
		errmsg = "setting configuration no failed";
		goto bad;
	}

	if (ueagle_getesi(sc, esi) != 0) {
		errmsg = "reading End System Identifier failed";
		goto bad;
	}

	printf("%s: End System Identifier: %02X:%02X:%02X:%02X:%02X:%02X\n",
	    USBDEVNAME(sc->sc_dev), esi[0], esi[1], esi[2], esi[3], esi[4],
	    esi[5]);

#if defined(__FreeBSD__)
	mtx_init(&sc->mtx, USBDEVNAME(sc->sc_dev), MTX_NETWORK_LOCK, MTX_DEF);
#endif

#if defined(__FreeBSD__)
	sc->ifatm.mib.device = ATM_DEVICE_UNKNOWN;
	bcopy(esi, sc->ifatm.mib.esi, 6);
	sc->ifatm.mib.serial = 0;
	sc->ifatm.mib.hw_version = 0;
	sc->ifatm.mib.sw_version = 0;
	sc->ifatm.mib.pcr = 0; /* XXX */
	sc->ifatm.mib.media = IFM_ATM_UNKNOWN;
	sc->ifatm.mib.vpi_bits = 8;
	sc->ifatm.mib.vci_bits = 16;
	sc->ifatm.mib.max_vccs = UEAGLE_MAXVCC;
	sc->ifatm.mib.max_vpcs = 0;
#endif

	ifp = GET_IFP(sc);
	ifp->if_softc = sc;
#if defined(__NetBSD__) || defined(__OpenBSD__)
	strncpy(ifp->if_xname, USBDEVNAME(sc->sc_dev), IFNAMSIZ);
#elif defined(__FreeBSD__)
	if_initname(ifp, "ueagle", USBDEVUNIT(sc->sc_dev));
#endif
	ifp->if_start = ueagle_start;
	ifp->if_ioctl = ueagle_ioctl;
#if defined(__FreeBSD__)
	ifp->if_init = ueagle_init;
#endif
	ifp->if_snd.ifq_maxlen = IFQ_MAXLEN;
	ifp->if_flags = IFF_SIMPLEX;
#ifndef UEAGLE_NO_MOZILLA_WORKAROUND
	/*
	 * XXX: ATM interfaces should be NBMA (Non-Broadcast Multiple Access)
	 * but mozilla does not behave correctly with such interfaces (output
	 * TTL is set to 1). As a workaround, we add the flags IFF_BROADCAST
	 * and IFF_MULTICAST to the interface flags.
	 */
	ifp->if_flags |= IFF_BROADCAST | IFF_MULTICAST;
#endif

#if defined(__NetBSD__) || defined(__OpenBSD__)
	if_attach(ifp);
#endif
	atm_ifattach(ifp);

	ifp->if_mtu = UEAGLE_IFMTU;

#if NBPFILTER > 0
#if defined(__FreeBSD__) || defined(__NetBSD__)
	bpfattach(ifp, DLT_RAW, 0);
#elif defined(__OpenBSD__)
	bpfattach(&ifp->if_bpf, ifp, DLT_RAW, 0);
#endif
#endif

	ueagle_crc_init();

	USB_ATTACH_SUCCESS_RETURN;

bad:	printf("%s: %s\n", USBDEVNAME(sc->sc_dev), errmsg);
	USB_ATTACH_ERROR_RETURN;
}

USB_DETACH(ueagle)
{
	USB_DETACH_START(ueagle, sc);
	struct ifnet *ifp = GET_IFP(sc);

	sc->gone = 1;

	ueagle_stop(sc);

#if NBPFILTER > 0
	bpfdetach(ifp);
#endif

#if defined(__FreeBSD__)
	atm_ifdetach(ifp);
#elif defined(__NetBSD__) || defined(__OpenBSD__)
	if_detach(ifp);
#endif
	return 0;
}

/* Retrieve the device End System Identifier (MAC) */
Static int
ueagle_getesi(struct ueagle_softc *sc, u_char *esi)
{
	usb_string_descriptor_t us;
	u_int16_t c;
	usbd_status error;
	int i;

#if defined(__OpenBSD__)
	error = usbd_get_string_desc(sc->sc_udev, UEAGLE_ESISTRING, 0, &us);
#else
	int len;
	error = usbd_get_string_desc(sc->sc_udev, UEAGLE_ESISTRING, 0, &us,
	    &len);
#endif
	if (error != 0)
		return error;

	if (us.bLength < (6 + 1) * 2)
		return 1;

	for (i = 0; i < 6 * 2; i++) {

		if ((c = UGETW(us.bString[i])) & 0xff00)
			return 1;

		if (i & 1)
			esi[i / 2] <<= 4;
		else
			esi[i / 2] = 0;

		if (c >= '0' && c <= '9')
			esi[i / 2] |= c - '0';
		else if (c >= 'a' && c <= 'f')
			esi[i / 2] |= c - 'a' + 10;
		else if (c >= 'A' && c <= 'F')
			esi[i / 2] |= c - 'A' + 10;
		else
			return 1;
	}

	return 0;
}

Static void
ueagle_idma_cb(usbd_xfer_handle xfer, usbd_private_handle priv,
    usbd_status status)
{
	struct ueagle_softc *sc = priv;

	if (status != USBD_NORMAL_COMPLETION) {
		if (status == USBD_STALLED)
			usbd_clear_endpoint_stall(sc->ep_idma);

		printf("%s: sending fw block failed: %s\n",
		    USBDEVNAME(sc->sc_dev), usbd_errstr(status));
	}

	usbd_free_xfer(xfer);
}

/*
 * The ueagle_loadpage() function must be called within a process context
 * under NetBSD and OpenBSD.
 */
Static void
#if defined(__NetBSD__) || defined(__OpenBSD__)
ueagle_loadpage(void *xsc)
{
	struct ueagle_softc *sc = xsc;
	u_int16_t pageno = sc->pageno;
	u_int16_t ovl = sc->ovl;
#elif defined(__FreeBSD__)
ueagle_loadpage(struct ueagle_softc *sc, u_int16_t pageno, u_int16_t ovl)
{
#endif
	usbd_xfer_handle xfer;
	block_info_t bi;
	const struct page *p;
	const struct block *b;
	int i;

	if (pageno >= ueaglefw_pagecount)
		goto bad1;

	p = &ueaglefw_pages[pageno];

	DPRINTF(("%s: sending %u blocks for fw page %u\n",
	    USBDEVNAME(sc->sc_dev), p->blockcount, pageno));

	USETW(bi.wHdr, UEAGLE_BIHDR);
	USETW(bi.wOvl, ovl);
	USETW(bi.wOvlOffset, ovl | 0x8000);
	for (i = 0, b = p->blocks; i < p->blockcount; i++, b++) {

		USETW(bi.wSize, b->size);
		USETW(bi.wAddress, b->address);
		USETW(bi.wLast, (i == p->blockcount - 1) ? 1 : 0);

		/* send block info through the IDMA pipe */
		if ((xfer = usbd_alloc_xfer(sc->sc_udev)) == NULL)
			goto bad2;

		usbd_setup_xfer(xfer, sc->ep_idma, sc, &bi, BLOCK_INFO_SIZE, 0,
		    UEAGLE_IDMA_TIMEOUT, ueagle_idma_cb);
		if (usbd_transfer(xfer) != USBD_IN_PROGRESS) {
			usbd_free_xfer(xfer);
			goto bad2;
		}

		/* send block data through the IDMA pipe */
		if ((xfer = usbd_alloc_xfer(sc->sc_udev)) == NULL)
			goto bad2;

		usbd_setup_xfer(xfer, sc->ep_idma, sc, (void *)b->data,
		    b->size, 0, UEAGLE_IDMA_TIMEOUT, ueagle_idma_cb);
		if (usbd_transfer(xfer) != USBD_IN_PROGRESS) {
			usbd_free_xfer(xfer);
			goto bad2;
		}
	}

	return;

bad2:	printf("%s: sending fw block %u failed\n", USBDEVNAME(sc->sc_dev), i);
	return;

bad1:	printf("%s: invalid fw page %u requested\n", USBDEVNAME(sc->sc_dev),
	    pageno);
}

#define ueagle_cr(sc, address, offset, callback) \
	ueagle_cmv(sc, MAKEFUNCTION(MEMACCESS, REQUESTREAD), address,	\
	    offset, 0, callback)

#define ueagle_cw(sc, address, offset, data, callback) \
	ueagle_cmv(sc, MAKEFUNCTION(MEMACCESS, REQUESTWRITE), address,	\
	    offset, data, callback)

Static void
ueagle_dispatch_cmv(struct ueagle_softc *sc, cmv_t *cmv)
{
	u_int32_t data;

	if (UGETW(cmv->wPreamble) != PREAMBLE)
		goto bad1;

	if (cmv->bDirection != MODEMTOHOST)
		goto bad1;

        if (cmv->bFunction == MAKEFUNCTION(ADSLDIRECTIVE, MODEMREADY)) {
		/* set CNTL0 to 1 and start sending configuration options */
		sc->stage = 0;
		ueagle_cw(sc, SA_CNTL, 0, 1, ueagle_optn_cb);
		return;
        }

	if (!sc->cmv_expected)
		goto bad2;

	if (UGETW (cmv->wIndex) != sc->cmv_idx ||
	    UGETDW(cmv->dwSymbolicAddress) != sc->cmv_address ||
	    UGETW (cmv->wOffsetAddress) != sc->cmv_offset)
		goto bad2;

	if (FUNCTION_TYPE(cmv->bFunction) != MEMACCESS)
		goto bad3;

	if (FUNCTION_SUBTYPE(cmv->bFunction) != REPLYREAD &&
	    FUNCTION_SUBTYPE(cmv->bFunction) != REPLYWRITE)
		goto bad3;

	data = UGETDW(cmv->dwData);
	data = data << 16 | data >> 16;

	sc->cmv_expected = 0;
	sc->cmv_cb(sc, CMV_NORMAL_COMPLETION, data);

	return;

bad3:	sc->cmv_expected = 0;
	sc->cmv_cb(sc, CMV_ERROR, 0);
	return;

bad2:	printf("%s: unexpected cmv received\n", USBDEVNAME(sc->sc_dev));
	return;

bad1:	printf("%s: invalid cmv received\n", USBDEVNAME(sc->sc_dev));
}

Static usbd_status
ueagle_request(struct ueagle_softc *sc, u_int16_t val, u_int16_t index,
    u_int16_t len, void *data)
{
	usb_device_request_t req;

	req.bmRequestType = UT_WRITE_VENDOR_DEVICE;
	req.bRequest = UCDC_SEND_ENCAPSULATED_COMMAND;
	USETW(req.wValue, val);
	USETW(req.wIndex, index);
	USETW(req.wLength, len);

	return usbd_do_request_async(sc->sc_udev, &req, data);
}

Static usbd_status
ueagle_cmv(struct ueagle_softc *sc, u_int8_t function, u_int32_t address,
    u_int16_t offset, u_int32_t data, cmv_cb_f cb)
{
	cmv_t cmv;

	USETW (cmv.wPreamble, PREAMBLE);
	cmv.bDirection = HOSTTOMODEM;
	cmv.bFunction = function;
	sc->cmv_idx++;
	USETW (cmv.wIndex, sc->cmv_idx);
	USETDW(cmv.dwSymbolicAddress, address);
	USETW (cmv.wOffsetAddress, offset);
	data = data >> 16 | data << 16;
	USETDW(cmv.dwData, data);

	sc->cmv_expected = 1;
	sc->cmv_address = address;
	sc->cmv_offset = offset;
	sc->cmv_cb = cb;

	/* XXX: should setup a timeout */

	return ueagle_request(sc, UEAGLE_SETBLOCK, UEAGLE_MPTXSTART, CMV_SIZE,
	    &cmv);
}

Static void
ueagle_stat_cb(struct ueagle_softc *sc, int status, u_int32_t data)
{
	if (status != CMV_NORMAL_COMPLETION) {
		printf("%s: failed to receive CMV ack: %d\n",
		    USBDEVNAME(sc->sc_dev), status);

		ueagle_startreset(sc);
		return;
	}

	switch (sc->stage) {
	case 0: /* STAT0 ack */
		sc->stats.phy.state = data;

		switch ((data >> 8) & 0xf) {
		case 0: /* not yet synchronized */
			DPRINTFN(3, ("%s: modem not yet synchronized\n",
			    USBDEVNAME(sc->sc_dev)));
			ueagle_sync(sc); /* retry */
			return;

		case 1: /* initialization */
			DPRINTFN(3, ("%s: modem initializing\n",
			    USBDEVNAME(sc->sc_dev)));
			ueagle_sync(sc); /* retry */
			return;

		case 2: /* operational */
			DPRINTFN(4, ("%s: modem operational\n",
			    USBDEVNAME(sc->sc_dev)));
			ueagle_cr(sc, SA_DIAG, 1, ueagle_stat_cb);
			break;

		default: /* fail ... */
			DPRINTFN(3, ("%s: modem synchronization failed\n",
			    USBDEVNAME(sc->sc_dev)));
			ueagle_startreset(sc);
			return;
		}
		break;

	case 1: /* DIAG1 ack */
		sc->stats.phy.flags = data;

		if (data & 0x10) {
			DPRINTF(("%s: delineation LOSS\n",
			    USBDEVNAME(sc->sc_dev)));
			ueagle_startreset(sc);
			return;
		}

		ueagle_cr(sc, SA_RATE, 0, ueagle_stat_cb);
		break;

	case 2: /* RATE0 ack */
		sc->stats.phy.dsrate = (data >> 16) * 32;
		sc->stats.phy.usrate = (data & 0xffff) * 32;
		ueagle_cr(sc, SA_DIAG, 3, ueagle_stat_cb);
		break;

	case 3:	/* DIAG3 ack */
		sc->stats.phy.cells_crc_errors = data;
		ueagle_cr(sc, SA_DIAG, 22, ueagle_stat_cb);
		break;

	case 4: /* DIAG22 ack */
		sc->stats.phy.dserror = data;
		ueagle_cr(sc, SA_DIAG, 23, ueagle_stat_cb);
		break;

	case 5: /* DIAG23 ack */
		sc->stats.phy.attenuation = (data & 0xff) / 2;
		ueagle_cr(sc, SA_DIAG, 25, ueagle_stat_cb);
		break;

	case 6: /* DIAG25 ack */
		sc->stats.phy.dsmargin = data;
		ueagle_cr(sc, SA_DIAG, 46, ueagle_stat_cb);
		break;

	case 7: /* DIAG46 ack */
		sc->stats.phy.userror = data;
		ueagle_cr(sc, SA_DIAG, 49, ueagle_stat_cb);
		break;

	case 8: /* DIAG49 ack */
		sc->stats.phy.usmargin = data;
		ueagle_cr(sc, SA_DIAG, 51, ueagle_stat_cb);
		break;

	case 9: /* DIAG51 ack */
		sc->stats.phy.rxflow = data;
		ueagle_cr(sc, SA_DIAG, 52, ueagle_stat_cb);
		break;

	case 10: /* DIAG52 ack */
		sc->stats.phy.txflow = data;
		ueagle_cr(sc, SA_DIAG, 54, ueagle_stat_cb);
		break;

	case 11: /* DIAG54 ack */
		sc->stats.phy.dsunc = data;
		ueagle_cr(sc, SA_DIAG, 58, ueagle_stat_cb);
		break;

	case 12: /* DIAG58 ack */
		sc->stats.phy.usunc = data;
		ueagle_cr(sc, SA_INFO, 8, ueagle_stat_cb);
		break;

	case 13: /* INFO8 ack */
		sc->stats.phy.vidco = data;
		ueagle_cr(sc, SA_INFO, 14, ueagle_stat_cb);
		break;

	case 14: /* INFO14 ack */
		sc->stats.phy.vidcpe = data;

		sc->stage = 0;
		usb_callout(sc->sc_ch, MS_TO_TICKS(1000), ueagle_stat, sc);

		/* wake up processes waiting for synchronization */
		wakeup(sc);
		return;
	}

	sc->stage++;
}

/* Start requesting statistics to the modem */
Static void
ueagle_stat(void *priv)
{
	struct ueagle_softc *sc = priv;

	DPRINTFN(4, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	ueagle_cr(sc, SA_STAT, 0, ueagle_stat_cb);
}

Static void
ueagle_cntl02_cb(struct ueagle_softc *sc, int status, u_int32_t data)
{
	if (status != CMV_NORMAL_COMPLETION) {
		printf("%s: failed to receive CMV ack: %d\n",
		    USBDEVNAME(sc->sc_dev), status);

		ueagle_startreset(sc);
		return;
	}

	sc->stage = 0;

	/* can't sleep since we are called from an interrupt context */
        usb_callout(sc->sc_ch, MS_TO_TICKS(10000), ueagle_stat, sc);
}

Static void
ueagle_sync(void *priv)
{
	struct ueagle_softc *sc = priv;

	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	ueagle_cw(sc, SA_CNTL, 0, 2, ueagle_cntl02_cb);
}

Static void
ueagle_optn_cb(struct ueagle_softc *sc, int status, u_int32_t data)
{
	if (status != CMV_NORMAL_COMPLETION) {
		printf("%s: failed to receive CMV ack: %d\n",
		    USBDEVNAME(sc->sc_dev), status);

		ueagle_startreset(sc);
		return;
	}

	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	switch (sc->stage) {
	case 0: /* CNTL0 ack */
		sc->stage = 1;
		ueagle_cw(sc, SA_OPTN, 0, UEAGLE_OPTN0, ueagle_optn_cb);
		break;

        case 1: /* OPTN0 ack */
		sc->stage = 2;
		ueagle_cw(sc, SA_OPTN, 2, UEAGLE_OPTN2, ueagle_optn_cb);
		break;

	case 2:
#ifdef UEAGLE_OPTN3
		sc->stage = 3;
		ueagle_cw(sc, SA_OPTN, 3, UEAGLE_OPTN3, ueagle_optn_cb);
		break;

#endif
	case 3:
#ifdef UEAGLE_OPTN4
		sc->stage = 4;
		ueagle_cw(sc, SA_OPTN, 4, UEAGLE_OPTN4, ueagle_optn_cb);
		break;

#endif
	case 4:
#ifdef UEAGLE_OPTN5
		sc->stage = 5;
		ueagle_cw(sc, SA_OPTN, 5, UEAGLE_OPTN5, ueagle_optn_cb);
		break;

#endif
	case 5:
#ifdef UEAGLE_OPTN6
		sc->stage = 6;
		ueagle_cw(sc, SA_OPTN, 6, UEAGLE_OPTN6, ueagle_optn_cb);
		break;

#endif
        case 6:
		sc->stage = 7;
		ueagle_cw(sc, SA_OPTN, 7, UEAGLE_OPTN7, ueagle_optn_cb);
		break;

	case 7:
#ifdef UEAGLE_OPTN15
		sc->stage = 8;
		ueagle_cw(sc, SA_OPTN, 15, UEAGLE_OPTN15, ueagle_optn_cb);
		break;

#endif
        case 8:
		ueagle_sync(sc);
		return;
        }
}

Static void
ueagle_bootstrap(void *priv)
{
	struct ueagle_softc *sc = priv;

	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	/*
	 * XXX: we should start a timeout in case the modem fails to send a
	 * modem ready CMV.
	 */

#if defined(__NetBSD__) || defined(__OpenBSD__)
	sc->pageno = 0;
	sc->ovl = 0;
	usb_add_task(sc->sc_udev, &sc->sc_task);
#elif defined(__FreeBSD__)
        ueagle_loadpage(sc, 0, 0);
#endif
}

/*-
 * 1) leave reset mode
 * 2) clear tx and rx mailboxes
 * 3) wait 1s before boot straping the firmware
 */
Static void
ueagle_endreset(void *priv)
{
	struct ueagle_softc *sc = priv;
	u_int16_t zero = 0; /* ;-) */

	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

        ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_ENDRESET, 0, NULL);
        ueagle_request(sc, UEAGLE_SET2183DATA, UEAGLE_MPTXMAILBOX, 2, &zero);
        ueagle_request(sc, UEAGLE_SET2183DATA, UEAGLE_MPRXMAILBOX, 2, &zero);
        ueagle_request(sc, UEAGLE_SET2183DATA, UEAGLE_SWAPMAILBOX, 2, &zero);

	/* can't sleep since we can be called from an interrupt context */
        usb_callout(sc->sc_ch, MS_TO_TICKS(1000), ueagle_bootstrap, sc);
}

/*-
 * 1) tell the modem that we want to boot in IDMA mode
 * 2) enter reset mode
 * 3) wait 200ms before leaving reset mode
 */
Static void
ueagle_startreset(struct ueagle_softc *sc)
{
	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

        ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_BOOTIDMA, 0, NULL);
        ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_STARTRESET, 0, NULL);

	/* can't sleep since we can be called from an interrupt context */
        usb_callout(sc->sc_ch, MS_TO_TICKS(200), ueagle_endreset, sc);
}

/* open the IDMA and interrupt pipes */
Static void
ueagle_boot(struct ueagle_softc *sc)
{
	usbd_interface_handle iface;
	char *errmsg;

	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	if (usbd_device2interface_handle(sc->sc_udev, UEAGLE_US_IFACE_NO,
	    &iface) != 0) {
		errmsg = "getting idma interface handle failed";
		goto bad;
	}

	if (usbd_open_pipe(iface, UEAGLE_IDMA_PIPE, USBD_EXCLUSIVE_USE,
	    &sc->ep_idma) != 0) {
		errmsg = "opening idma pipe failed";
		goto bad;
	}

	if (usbd_device2interface_handle(sc->sc_udev, UEAGLE_INTR_IFACE_NO,
	    &iface) != 0) {
		errmsg = "getting interrupt interface handle failed";
		goto bad;
	}

	if ((sc->ibuf = malloc(INTR_PKT_SIZE, M_USBDEV, M_NOWAIT)) == NULL) {
		errmsg = "allocating interrupt buffer failed";
		goto bad;
	}

	if (usbd_open_pipe_intr(iface, UEAGLE_INTR_PIPE, USBD_SHORT_XFER_OK,
	    &sc->ep_intr, sc, sc->ibuf, INTR_PKT_SIZE, ueagle_intr,
	    UEAGLE_INTR_INTERVAL) != 0) {
		errmsg = "opening interrupt pipe failed";
		goto bad;
	}

	ueagle_startreset(sc);

	return;

bad:	printf("%s: %s\n", USBDEVNAME(sc->sc_dev), errmsg);
	ueagle_stop(sc);
}

Static void
ueagle_intr(usbd_xfer_handle xfer, usbd_private_handle priv, usbd_status status)
{
	struct ueagle_softc *sc = priv;
	intr_pkt_t *intr;

	if (status != USBD_NORMAL_COMPLETION) {
		if (status == USBD_STALLED)
			usbd_clear_endpoint_stall_async(sc->ep_intr);

		return;
	}

	intr = sc->ibuf;

	switch (UGETW(intr->wInterrupt)) {
	case INT_LOADSWAPPAGE:
#if defined(__NetBSD__) || defined(__OpenBSD__)
		sc->pageno = intr->bSwapPageNo;
		sc->ovl = intr->bOvl >> 4 | intr->bOvl << 4;
		usb_add_task(sc->sc_udev, &sc->sc_task);
#elif defined(__FreeBSD__)
		ueagle_loadpage(sc, intr->bSwapPageNo,
		    intr->bOvl >> 4 | intr->bOvl << 4);
#endif
		break;

	case INT_INCOMINGCMV:
		ueagle_dispatch_cmv(sc, &intr->u.s2.cmv);
		break;

	default:
		printf("%s: unknown intr %u\n", USBDEVNAME(sc->sc_dev),
		    UGETW(intr->wInterrupt));
	}
}

Static u_int32_t ueagle_crc_tbl[256];
Static int ueagle_crc_inited = 0;

Static void
ueagle_crc_init(void)
{
	u_int32_t c;
	int n, k;

	if (ueagle_crc_inited)
		return;

	for (n = 0; n < 256; n++) {
		c = ((u_int32_t)n) << 24;
		for (k = 0; k < 8; k++) {
			if (c & 0x80000000)
				c = CRC_POLYNOM ^ (c << 1);
			else
				c <<= 1;
		}
		ueagle_crc_tbl[n] = c;
	}

	ueagle_crc_inited = 1;
}

/* XXX: could be optimized */
Static u_int32_t
ueagle_crc_update(u_int32_t crc, u_int8_t *buf, int len)
{
	for (; len != 0; len--, buf++)
		crc = ueagle_crc_tbl[((crc >> 24) ^ *buf) & 0xff] ^ (crc << 8);

	return crc;
}

Static struct ueagle_vcc *
ueagle_find_vcc(struct ueagle_softc *sc, u_int8_t vpi, u_int16_t vci)
{
	struct ueagle_vcc *vcc;

	if ((vcc = sc->vccs[0]) == NULL)
		return NULL;

	if (vcc->param.vpi != vpi || vcc->param.vci != vci)
		return NULL;

	return vcc;
}

Static void
ueagle_record(struct ueagle_softc *sc, uint8_t *cell)
{
	struct ueagle_vcc *vcc;
	struct ifnet *ifp;
	struct mbuf *m;
	u_int32_t crc;
	u_int16_t pdulen, totlen;

	sc->stats.atm.cells_received++;

	vcc = ueagle_find_vcc(sc, ATM_CH_GETVPI(cell), ATM_CH_GETVCI(cell));
	if (vcc == NULL) {
		sc->stats.atm.vcc_no_conn++;
		return;
	}

	if (vcc->drop) {
		if (ATM_CH_ISLASTCELL(cell)) {
			vcc->drop = 0; /* leave drop mode */
			sc->stats.atm.cspdus_dropped++;
		}

		sc->stats.atm.cells_dropped++;
		return;
	}

	if (vcc->m_head == NULL) {

		MGETHDR(m, M_DONTWAIT, MT_DATA);
		if (m == NULL) {
			vcc->drop = 1; /* enter drop mode */
			return;
		}

		MCLGET(m, M_DONTWAIT);
		if (!(m->m_flags & M_EXT)) {
			vcc->drop = 1; /* enter drop mode */
			m_freem(m);
			return;
		}

		vcc->m_head = m;
		vcc->dst = mtod(m, u_int8_t *);
		vcc->limit = vcc->dst + MCLBYTES - ATM_CELL_PAYLOAD_SIZE;
	}

	if (vcc->dst > vcc->limit) {
		vcc->drop = 1; /* enter drop mode */
		sc->stats.atm.cells_dropped++;
		goto bad;
	}

	bcopy(cell + ATM_CELL_HEADER_SIZE, vcc->dst, ATM_CELL_PAYLOAD_SIZE);
	vcc->dst += ATM_CELL_PAYLOAD_SIZE;

	if (!ATM_CH_ISLASTCELL(cell))
		return;

	/* Handle the last cell of the AAL5 CPCS-PDU */

	m = vcc->m_head;

	totlen = vcc->dst - mtod(m, u_int8_t *);
	pdulen = AAL5_TR_GETPDULEN(cell);

	if (totlen < pdulen + AAL5_TRAILER_SIZE) {
		sc->stats.atm.cspdus_dropped++;
		goto bad;
	}

	if (totlen >= pdulen + ATM_CELL_PAYLOAD_SIZE + AAL5_TRAILER_SIZE) {
		sc->stats.atm.cspdus_dropped++;
		goto bad;
	}

	crc = ueagle_crc_update(CRC_INITIAL, mtod(m, u_int8_t *), totlen);
	if (crc != CRC_MAGIC) {
		sc->stats.atm.cspdus_crc_errors++;
		goto bad;
	}

	/* received AAL5 CPCS-PDU seems correct, finalize mbuf */
	ifp = GET_IFP(sc);
	m->m_pkthdr.rcvif = ifp;
	m->m_pkthdr.len = m->m_len = pdulen;

	sc->stats.atm.cspdus_received++;

#if NBPFILTER > 0
	BPF_MTAP(ifp, m);
#endif
	atm_input(ifp, &vcc->aph, m, vcc->rxhand);
	vcc->m_head = NULL;

	return;

bad:	/* XXX: vcc->dst = mtod(m, u_int8_t *); instead of freeing mbuf? */
	m_freem(vcc->m_head);
	vcc->m_head = NULL;
}

Static void
ueagle_rx_cb(usbd_xfer_handle xfer, usbd_private_handle priv,
    usbd_status status)
{
	struct isoreq *req = priv;
	struct ueagle_softc *sc = req->sc;
	u_int32_t count;
	u_char *p;
	int i;

	if (status == USBD_CANCELLED)
		return;

	for (i = 0; i < UEAGLE_NISOFRMS; i++) {
		count = req->frlengths[i];
		p = req->offsets[i];

		while (count >= ATM_CELL_SIZE) {
			ueagle_record(sc, p);
			p += ATM_CELL_SIZE;
			count -= ATM_CELL_SIZE;
		}
#ifdef DIAGNOSTIC
		if (count > 0)
			printf("%s: incomplete ATM cell received\n",
			    USBDEVNAME(sc->sc_dev));
#endif

		req->frlengths[i] = sc->isize;
	}

	usbd_setup_isoc_xfer(req->xfer, sc->ep_rx, req, req->frlengths,
	    UEAGLE_NISOFRMS, USBD_NO_COPY, ueagle_rx_cb);
	(void)usbd_transfer(xfer);
}

Static void
ueagle_tx_cb(usbd_xfer_handle xfer, usbd_private_handle priv,
    usbd_status status)
{
	struct txreq *req = priv;
	struct ueagle_softc *sc = req->sc;
	struct ifnet *ifp = GET_IFP(sc);
	int s;

	s = splnet();

	if (status != USBD_NORMAL_COMPLETION) {
		if (status == USBD_STALLED)
			usbd_clear_endpoint_stall(sc->ep_tx);

		ifp->if_oerrors++;
	}

	ifp->if_flags &= ~IFF_OACTIVE;

	ueagle_start(ifp);

	splx(s);
}

Static int
ueagle_send(struct ueagle_softc *sc, struct mbuf *m_head, int idx)
{
	struct atm_pseudohdr *aph;
	struct ueagle_vcc *vcc;
	struct txreq *req;
	struct mbuf *m;
	u_int8_t *src, *dst;
	u_int32_t crc;
	int n, cellleft, mleft;

	aph = mtod(m_head, struct atm_pseudohdr *);

	vcc = ueagle_find_vcc(sc, ATM_PH_VPI(aph), ATM_PH_VCI(aph));
	if (vcc == NULL)
		return EIO;

	m_adj(m_head, sizeof (struct atm_pseudohdr));

	req = &sc->txreqs[idx];

	dst = req->buf;
	cellleft = 0;
	crc = CRC_INITIAL;

	for (m = m_head; m != NULL; m = m->m_next) {
		src = mtod(m, u_int8_t *);
		mleft = m->m_len;

		crc = ueagle_crc_update(crc, src, mleft);

		if (cellleft != 0) {
			n = min(mleft, cellleft);

			bcopy(src, dst, n);
			dst += n;
			src += n;
			cellleft -= n;
			mleft -= n;
		}

		while (mleft >= ATM_CELL_PAYLOAD_SIZE) {
			bcopy(vcc->ch, dst, ATM_CELL_HEADER_SIZE);
			dst += ATM_CELL_HEADER_SIZE;
			bcopy(src, dst, ATM_CELL_PAYLOAD_SIZE);
			dst += ATM_CELL_PAYLOAD_SIZE;
			src += ATM_CELL_PAYLOAD_SIZE;
			mleft -= ATM_CELL_PAYLOAD_SIZE;
			sc->stats.atm.cells_transmitted++;
		}

		if (mleft != 0) {
			bcopy(vcc->ch, dst, ATM_CELL_HEADER_SIZE);
			dst += ATM_CELL_HEADER_SIZE;
			bcopy(src, dst, mleft);
			dst += mleft;
			cellleft = ATM_CELL_PAYLOAD_SIZE - mleft;
			sc->stats.atm.cells_transmitted++;
		}
	}

	/*
	 * If there is not enough space to put the AAL5 trailer into this cell,
	 * pad the content of this cell with zeros and create a new cell which
	 * will contain no data except the AAL5 trailer itself.
	 */
	if (cellleft < AAL5_TRAILER_SIZE) {
		bzero(dst, cellleft);
		crc = ueagle_crc_update(crc, dst, cellleft);
		dst += cellleft;

		bcopy(vcc->ch, dst, ATM_CELL_HEADER_SIZE);
		dst += ATM_CELL_HEADER_SIZE;
		cellleft = ATM_CELL_PAYLOAD_SIZE;
		sc->stats.atm.cells_transmitted++;
	}

	/* Fill the AAL5 CPCS-PDU trailer */

	bzero(dst, cellleft - AAL5_TRAILER_SIZE);

	/* src now points to the beginning of the last cell */
	src = dst + cellleft - ATM_CELL_SIZE;
	ATM_CH_SETPTFLAGS(src, 1);

	AAL5_TR_SETCPSUU(src, 0);
	AAL5_TR_SETCPI(src, 0);
	AAL5_TR_SETPDULEN(src, m_head->m_pkthdr.len);

	crc = ~ueagle_crc_update(crc, dst, cellleft - 4);
	AAL5_TR_SETCRC(src, crc);

	sc->stats.atm.cspdus_transmitted++;

	usbd_setup_xfer(req->xfer, sc->ep_tx, req, req->buf,
	    dst + cellleft - req->buf, USBD_FORCE_SHORT_XFER | USBD_NO_COPY,
	    UEAGLE_TX_TIMEOUT, ueagle_tx_cb);
	if (usbd_transfer(req->xfer) != USBD_IN_PROGRESS)
		return EIO;

	return 0;
}

/*
 * A full PDU is ready for transmission. Segment it into multiple ATM cells
 * and send it through the USB bulk tx pipe.
 */
Static void
ueagle_start(struct ifnet *ifp)
{
	struct ueagle_softc *sc = ifp->if_softc;
	struct mbuf *m_head;
	int s;

	s = splnet();

	IF_DEQUEUE(&ifp->if_snd, m_head);
	if (m_head == NULL) {
		splx(s);
		return;
	}

	if (ueagle_send(sc, m_head, 0) != 0) {
		ifp->if_flags |= IFF_OACTIVE;
		m_freem(m_head);
		splx(s);
		return;
	}

#if NBPFILTER > 0
	BPF_MTAP(ifp, m_head);
#endif

	ifp->if_flags |= IFF_OACTIVE;

	m_freem(m_head);

	splx(s);
}

Static int
ueagle_tx_init(struct ueagle_softc *sc)
{
	usbd_interface_handle iface;
	struct txreq *req;
	char *errmsg;
	int i;

	DPRINTFN(2, ("%s: tx initialization\n", USBDEVNAME(sc->sc_dev)));

	if (usbd_device2interface_handle(sc->sc_udev, UEAGLE_US_IFACE_NO,
	    &iface) != 0) {
		errmsg = "getting tx interface handle failed";
		goto bad;
	}

	if (usbd_open_pipe(iface, UEAGLE_TX_PIPE, USBD_EXCLUSIVE_USE,
	    &sc->ep_tx) != 0) {
		errmsg = "opening tx pipe failed";
		goto bad;
	}

	for (i = 0, req = sc->txreqs; i < UEAGLE_TX_LIST_CNT; i++, req++) {
		req->sc = sc;

		if ((req->xfer = usbd_alloc_xfer(sc->sc_udev)) == NULL) {
			errmsg = "allocating tx xfer failed";
			goto bad;
		}

		req->buf = usbd_alloc_buffer(req->xfer, UEAGLE_TXBUFLEN);
		if (req->buf == NULL) {
			errmsg = "allocating tx xfer buffer failed";
			goto bad;
		}
	}

	return 0;

bad:	printf("%s: %s\n", USBDEVNAME(sc->sc_dev), errmsg);
	ueagle_tx_stop(sc);

	return 1;
}

Static void
ueagle_tx_stop(struct ueagle_softc *sc)
{
	int i;

	if (sc->ep_tx != NULL) {
		usbd_abort_pipe(sc->ep_tx);
		usbd_close_pipe(sc->ep_tx);
		sc->ep_tx = NULL;
	}

	for (i = 0; i < UEAGLE_TX_LIST_CNT && sc->txreqs[i].xfer != NULL; i++)
		usbd_free_xfer(sc->txreqs[i].xfer);
	sc->txreqs[0].xfer = NULL;
}

Static int
ueagle_rx_init(struct ueagle_softc *sc)
{
	usbd_interface_handle iface;
	usb_endpoint_descriptor_t *edesc;
	struct isoreq *req;
	u_char *buf;
	char *errmsg;
	int i, j;

	DPRINTFN(2, ("%s: rx initialization\n", USBDEVNAME(sc->sc_dev)));

	if (usbd_device2interface_handle(sc->sc_udev, UEAGLE_DS_IFACE_NO,
	    &iface) != 0) {
		errmsg = "getting rx interface handle failed";
		goto bad;
	}

	/* XXX: sould depend on downrate */
	if (usbd_set_interface(iface, 6) != 0) {
		errmsg = "setting alternative rx interface failed";
		goto bad;
	}

	if ((edesc = usbd_get_endpoint_descriptor(iface, 0x88)) == NULL) {
		errmsg = "getting rx endpoint decriptor failed";
		goto bad;
	}

	sc->isize = UGETW(edesc->wMaxPacketSize);

	if (usbd_open_pipe(iface, UEAGLE_RX_PIPE, USBD_EXCLUSIVE_USE,
	    &sc->ep_rx) != 0) {
		errmsg = "opening rx pipe failed";
		goto bad;
	}

	for (i = 0, req = sc->isoreqs; i < UEAGLE_NISOREQS; i++, req++) {
		req->sc = sc;

		if ((req->xfer = usbd_alloc_xfer(sc->sc_udev)) == NULL) {
			errmsg = "allocating rx xfer failed";
			goto bad;
		}

		buf = usbd_alloc_buffer(req->xfer, sc->isize * UEAGLE_NISOFRMS);
		if (buf == NULL) {
			errmsg = "allocating rx xfer buffer failed";
			goto bad;
		}

		for (j = 0; j < UEAGLE_NISOFRMS; j++) {
			req->frlengths[j] = sc->isize;
			req->offsets[j] = buf + j * sc->isize;
		}

		usbd_setup_isoc_xfer(req->xfer, sc->ep_rx, req, req->frlengths,
		    UEAGLE_NISOFRMS, USBD_NO_COPY, ueagle_rx_cb);
		(void)usbd_transfer(req->xfer);
	}

	return 0;

bad:	printf("%s: %s\n", USBDEVNAME(sc->sc_dev), errmsg);
	ueagle_rx_stop(sc);

	return 1;
}

Static void
ueagle_rx_stop(struct ueagle_softc *sc)
{
	int i;

	if (sc->ep_rx != NULL) {
		usbd_abort_pipe(sc->ep_rx);
		usbd_close_pipe(sc->ep_rx);
		sc->ep_rx = NULL;
	}

	for (i = 0; i < UEAGLE_NISOREQS && sc->isoreqs[i].xfer != NULL; i++)
		usbd_free_xfer(sc->isoreqs[i].xfer);
	sc->isoreqs[0].xfer = NULL;
}

Static void
ueagle_init(void *priv)
{
	struct ueagle_softc *sc = priv;
	struct ifnet *ifp = GET_IFP(sc);

	if (((sc->stats.phy.state >> 8) & 0xf) == 2)
		return;

	ueagle_stop(sc);

	ueagle_boot(sc);

	tsleep(sc, PZERO | PCATCH, "tsleep", 0);

	if (ueagle_tx_init(sc) != 0)
		return;

	if (ueagle_rx_init(sc) != 0) {
		ueagle_tx_stop(sc);
		return;
	}

        ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_LOOPBACKOFF, 0, NULL);

	/* set running flag and clear output active flag */
	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;
}

Static void
ueagle_stop(struct ueagle_softc *sc)
{
	struct ifnet *ifp = GET_IFP(sc);

#if defined(__NetBSD__) || defined(__OpenBSD__)
	usb_rem_task(sc->sc_udev, &sc->sc_task);
#endif
	/* stop any pending boot process */
	usb_uncallout(sc->sc_ch, ueagle_bootstrap, sc);
	usb_uncallout(sc->sc_ch, ueagle_endreset, sc);
	usb_uncallout(sc->sc_ch, ueagle_stat, sc);

        if (!sc->gone)
		ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_LOOPBACKON, 0, NULL);

	ueagle_rx_stop(sc);
	ueagle_tx_stop(sc);

	if (sc->ep_idma != NULL) {
		usbd_abort_pipe(sc->ep_idma);
		usbd_close_pipe(sc->ep_idma);
		sc->ep_idma = NULL;
	}

	if (sc->ep_intr != NULL) {
		usbd_abort_pipe(sc->ep_intr);
		usbd_close_pipe(sc->ep_intr);
		sc->ep_intr = NULL;
	}

	if (sc->ibuf != NULL) {
		free(sc->ibuf, M_USBDEV);
		sc->ibuf = NULL;
	}

	/* reset statistics */
	bzero(&sc->stats, sizeof (struct ueagle_stats));

	/* clear running and output active flags */
	ifp->if_flags &= ~(IFF_RUNNING | IFF_OACTIVE);

}

Static int
ueagle_open_vcc(struct ueagle_softc *sc, caddr_t data)
{
	struct ueagle_vcc *vcc;
#if defined(__FreeBSD__) && __FreeBSD_version >= 502000
	struct atmio_openvcc *aop = (struct atmio_openvcc *)data;
#else
	struct atm_pseudoioctl *api = (struct atm_pseudoioctl *)data;
#endif

	if (sc->vccs[0] != NULL)
		return EBUSY;

	DPRINTF(("%s: opening ATM virtual circuit\n", USBDEVNAME(sc->sc_dev)));

#if defined(__FreeBSD__) && __FreeBSD_version >= 502000
	if (aop->param.aal != ATMIO_AAL_5)
		return EINVAL;
#endif

	vcc = malloc(sizeof (struct ueagle_vcc), M_DEVBUF, M_NOWAIT);
	if (vcc == NULL)
		return ENOMEM;

#if defined(__FreeBSD__) && __FreeBSD_version >= 502000
	vcc->param = aop->param;
	vcc->rxhand = aop->rxhand;

	/* pre-calculate pseudo-header */
	ATM_PH_FLAGS(&vcc->aph) = ATM_PH_AAL5;
	ATM_PH_VPI(&vcc->aph) = vcc->param.vpi;
	ATM_PH_SETVCI(&vcc->aph, vcc->param.vci);
#else
	vcc->param.flags = ATM_PH_FLAGS(&api->aph);
	vcc->param.vpi = ATM_PH_VPI(&api->aph);
	vcc->param.vci = ATM_PH_VCI(&api->aph);
	vcc->rxhand = api->rxhand;

	/* store pseudo-header */
	vcc->aph = api->aph;
#endif
	/* pre-calculate cell headers (HEC field will be set by the modem) */
	ATM_CH_FILL(vcc->ch, 0, vcc->param.vpi, vcc->param.vci, 0, 0, 0);

	vcc->m_head = NULL;
	vcc->drop = 0;

	sc->vccs[0] = vcc;
	sc->open_vccs++;

	return 0;
}

Static int
ueagle_close_vcc(struct ueagle_softc *sc, caddr_t data)
{
	if (sc->vccs[0] == NULL)
		return ENOENT;

	DPRINTF(("%s: closing ATM virtual circuit\n", USBDEVNAME(sc->sc_dev)));

	free(sc->vccs[0], M_DEVBUF);
	sc->vccs[0] = NULL;
	sc->open_vccs--;

	return 0;
}

Static int
ueagle_ioctl(struct ifnet *ifp, u_long cmd, caddr_t data)
{
	struct ueagle_softc *sc = ifp->if_softc;
	struct ifaddr *ifa;
	struct ifreq *ifr;
#if defined(__FreeBSD__) && __FreeBSD_version >= 502000
	struct atmio_vcctable *vtab;
#endif
	usb_config_descriptor_t *cdesc;
	int len;
	struct iovec iov;
	struct uio uio;

	switch (cmd) {
#if defined(__FreeBSD__) && __FreeBSD_version >= 502000
	case SIOCATMOPENVCC:
		return ueagle_open_vcc(sc, data);

	case SIOCATMCLOSEVCC:
		return ueagle_close_vcc(sc, data);

	case SIOCATMGVCCS:
		ifr = (struct ifreq *)data;
		vtab = atm_getvccs((struct atmio_vcc **)sc->vccs,
		    UEAGLE_MAXVCC, sc->open_vccs, &sc->mtx, 1);
		if (vtab == NULL)
			return ENOMEM;

		if (copyout(vtab, ifr->ifr_data, sizeof (*vtab) +
		    vtab->count * sizeof (struct atmio_vcc)) != 0)
			return EFAULT;

		free(vtab, M_DEVBUF);
		break;

	case SIOCATMGETVCCS:	/* used by netgraph */
		vtab = atm_getvccs((struct atmio_vcc **)sc->vccs,
		    UEAGLE_MAXVCC, sc->open_vccs, &sc->mtx, 0);
		if (vtab == NULL)
			return ENOMEM;

		*(void **)data = vtab;
		break;
#else
	case SIOCATMENA:
		return ueagle_open_vcc(sc, data);

	case SIOCATMDIS:
		return ueagle_close_vcc(sc, data);
#endif

	case SIOCGATMSTATS:
		ifr = (struct ifreq *)data;

		return copyout(&sc->stats, ifr->ifr_data,
		    sizeof (struct ueagle_stats));

	case SIOCGUSBDEVICE:
		ifr = (struct ifreq *)data;

		return copyout(usbd_get_device_descriptor(sc->sc_udev),
		    ifr->ifr_data, sizeof (usb_device_descriptor_t));

	case SIOCGUSBCONFIG:
		ifr = (struct ifreq *)data;

		return copyout(usbd_get_config_descriptor(sc->sc_udev),
		    ifr->ifr_data, sizeof (struct usb_config_desc));

	case SIOCGUSBFULLDESC:
		ifr = (struct ifreq *)data;

		cdesc = usbd_get_config_descriptor(sc->sc_udev);
		len = UGETW(cdesc->wTotalLength);

		iov.iov_base = (caddr_t)ifr->ifr_data;
		iov.iov_len = len;
		uio.uio_iov = &iov;
		uio.uio_iovcnt = 1;
		uio.uio_resid = len;
		uio.uio_offset = 0;
		uio.uio_segflg = UIO_USERSPACE;
		uio.uio_rw = UIO_READ;
#if defined(__FreeBSD__)
		uio.uio_td = curthread;
#else
		uio.uio_procp = curproc;
#endif
		return uiomove((void *)cdesc, len, &uio);

	case SIOCSIFADDR:
		ifa = (struct ifaddr *)data;

		ifp->if_flags |= IFF_UP;
		if (!(ifp->if_flags & IFF_RUNNING))
			ueagle_init(sc);

#ifdef INET
		ifa->ifa_rtrequest = atm_rtrequest;
#endif
		break;

	case SIOCSIFFLAGS:
		if (ifp->if_flags & IFF_UP) {
			if (!(ifp->if_flags & IFF_RUNNING))
				ueagle_init(sc);
		} else {
			if (ifp->if_flags & IFF_RUNNING)
				ueagle_stop(sc);
		}
		break;

	case SIOCSIFMTU:
		ifr = (struct ifreq *)data;
		if (ifr->ifr_mtu > UEAGLE_IFMTU)
			return EINVAL;

		ifp->if_mtu = ifr->ifr_mtu;
		break;

	default:
		return EINVAL;
	}

	return 0;
}

#if defined(__NetBSD__) || defined(__OpenBSD__)
int
ueagle_activate(device_ptr_t self, enum devact act)
{
	switch (act) {
	case DVACT_ACTIVATE:
		return EOPNOTSUPP;

	case DVACT_DEACTIVATE:
		break;
	}

	return 0;
}
#elif defined(__FreeBSD__)
DRIVER_MODULE(ueagle, uhub, ueagle_driver, ueagle_devclass, usbd_driver_load, 0);
#endif
