/*	$Id: eaglectl.c,v 1.20 2004/08/29 19:09:54 damien Exp $	*/

/*-
 * Copyright (c) 2003, 2004
 *	Damien Bergamini <damien.bergamini@free.fr>. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
#if defined(__RCSID)
__RCSID("$Id: eaglectl.c,v 1.20 2004/08/29 19:09:54 damien Exp $");
#else
#ifndef lint
static char rcsid[] = "$Id: eaglectl.c,v 1.20 2004/08/29 19:09:54 damien Exp $";
#endif /* not lint */
#endif

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <net/if.h>

#include <ctype.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#define SIOCGATMSTATS	_IOWR('i', 138, struct ifreq)

struct eagle_stats {
	struct {
		u_int32_t	state;
		u_int32_t	flags;
		u_int32_t	vidcpe;
		u_int32_t	vidco;
		u_int32_t	dsrate;
		u_int32_t	usrate;
		u_int32_t	dserror;
		u_int32_t	userror;
		u_int32_t	dsunc;
		u_int32_t	usunc;
		u_int32_t	txflow;
		u_int32_t	rxflow;
		u_int32_t	attenuation;
		u_int32_t	dsmargin;
		u_int32_t	usmargin;
		u_int32_t	cells_crc_errors;
	} phy;

	struct {
		u_int32_t	cells_transmitted;
		u_int32_t	cells_received;
		u_int32_t	cells_dropped;
		u_int32_t	vcc_no_conn;
		u_int32_t	cspdus_transmitted;
		u_int32_t	cspdus_received;
		u_int32_t	cspdus_crc_errors;
		u_int32_t	cspdus_dropped;
	} atm;
};

extern char *optarg;
extern int optind;
extern int optopt;
extern int opterr;
extern int optreset;

static void usage(void);
static int do_req(char *, unsigned long, void *);
static void get_statistics(char *);

int
main(int argc, char **argv)
{
	int ch;
	char *iface;

        opterr = 0;
        ch = getopt(argc, argv, "i:");
        if (ch == 'i') {
                iface = optarg;
        } else {
                if (argc > 1 && argv[1][0] != '-') {
                        iface = argv[1];
                        optind = 2;
                } else {
                        iface = "ueagle0";
                        optind = 1;
                }
                optreset = 1;
        }
        opterr = 1;

        get_statistics(iface);

	return EX_OK;
}

static void
usage(void)
{
	extern char *__progname;

	(void)fprintf(stderr, "usage: %s -i iface\n", __progname);
	exit(EX_USAGE);
}

static int
do_req(char *iface, unsigned long req, void *data)
{
        int s;
        struct ifreq ifr;
        int error;

        if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
                err(EX_OSERR, "Can't create socket");

        memset(&ifr, 0, sizeof(ifr));
        strncpy(ifr.ifr_name, iface, sizeof(ifr.ifr_name));
        ifr.ifr_data = data;

        error = ioctl(s, req, &ifr);

        (void)close(s);

        return error;
}

static void
get_statistics(char *iface)
{
	struct eagle_stats stats;

	if (do_req(iface, SIOCGATMSTATS, &stats) == -1)
		err(EX_OSERR, "Can't retrieve statistics");

	printf("DSL Delineation    : %s\n",
		(stats.phy.flags & 0x20) ? "ERROR" :
		(stats.phy.flags & 0x10) ? "LOSS" : "GOOD");
	printf("DSL DownRate       : %uKb/s\n", stats.phy.dsrate);
	printf("DSL UpRate         : %uKb/s\n", stats.phy.usrate);
	printf("DSL Ds Margin      : %udB\n",   stats.phy.dsmargin);
	printf("DSL Us Margin      : %udB\n",   stats.phy.usmargin);
	printf("DSL Attenuation    : %udB\n",   stats.phy.attenuation);
	printf("DSL VID-CPE        : %u\n", stats.phy.vidcpe);
	printf("DSL VID-CO         : %u\n", stats.phy.vidco);
	printf("DSL Tx Flow        : %u\n", stats.phy.txflow);
	printf("DSL Rx Flow        : %u\n", stats.phy.rxflow);
	printf("DSL Tx Flow errors : %u\n", stats.phy.userror);
	printf("DSL Rx Flow errors : %u\n", stats.phy.dserror);
	printf("Cells transmitted  : %u\n", stats.atm.cells_transmitted);
	printf("Cells received     : %u\n", stats.atm.cells_received);
	printf("Cells HEC errors   : %u\n", stats.phy.cells_crc_errors);
	printf("Cells dropped      : %u\n", stats.atm.cells_dropped);
	printf("Bad VPI/VCI values : %u\n", stats.atm.vcc_no_conn);
	printf("CSPDUs transmitted : %u\n", stats.atm.cspdus_transmitted);
	printf("CSPDUs received	   : %u\n", stats.atm.cspdus_received);
	printf("CSPDUs CRC errors  : %u\n", stats.atm.cspdus_crc_errors);
	printf("CSPDUs dropped     : %u\n", stats.atm.cspdus_dropped);
}
