/*	$Id: eagleflash.c,v 1.11 2004/02/26 08:34:01 damien Exp $	*/

/*-
 * Copyright (c) 2003, 2004
 *	Damien Bergamini <damien.bergamini@free.fr>. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
#if defined(__RCSID)
__RCSID("$Id: eagleflash.c,v 1.11 2004/02/26 08:34:01 damien Exp $");
#else
#ifndef lint
static char rcsid[] = "$Id: eagleflash.c,v 1.11 2004/02/26 08:34:01 damien Exp $";
#endif /* not lint */
#endif

#include <sys/types.h>
#include <sys/ioctl.h>

#include <dev/usb/usb.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "firmware.h"

#define USB_VENDOR_ADI			0x1110
#define USB_PRODUCT_ADI_EAGLEI		0x9010
#define USB_PRODUCT_ADI_EAGLEII		0x9022
#define USB_PRODUCT_ADI_EAGLEIIC	0x9024

#define USB_VENDOR_USR			0x0baf
#define USB_PRODUCT_USR_MILLER_A	0x00f2
#define USB_PRODUCT_USR_MILLER_B	0x00fa
#define USB_PRODUCT_USR_HEINEKEN_A	0x00f6
#define USB_PRODUCT_USR_HEINEKEN_B	0x00f8

#define ANCHOR_LOAD_INTERNAL	0xa0
#define ANCHOR_CPUCS_REG	0x7f92
#define ANCHOR_RESET		0x01

extern char *optarg;
extern int optind;

struct ugen_info {
	u_int16_t vendor;
	u_int16_t product;
	struct ezdata *firmware;
};

static const struct ugen_info tbl[] = {
	/* Analog Devices pre-firmware devices */
	{ USB_VENDOR_ADI, USB_PRODUCT_ADI_EAGLEI,     fw_eagleI  },
	{ USB_VENDOR_ADI, USB_PRODUCT_ADI_EAGLEII,    fw_eagleII },
	{ USB_VENDOR_ADI, USB_PRODUCT_ADI_EAGLEIIC,   fw_eagleII },

	/* U.S. Robotics pre-firmware devices */
	{ USB_VENDOR_USR, USB_PRODUCT_USR_MILLER_A,   fw_eagleI  },
	{ USB_VENDOR_USR, USB_PRODUCT_USR_MILLER_B,   fw_eagleI  },
	{ USB_VENDOR_USR, USB_PRODUCT_USR_HEINEKEN_A, fw_eagleI  },
	{ USB_VENDOR_USR, USB_PRODUCT_USR_HEINEKEN_B, fw_eagleI  },

	{ 0, 0, NULL }
};

static void usage(void);
static struct ezdata *get_firmware(struct usb_device_info *);
static void flash(int, struct ezdata *);
static void scan_controller(int);
static void scan_all(void);

int
main(int argc, char **argv)
{
	int ch;
	int fd;
	char *devname = NULL;
	struct usb_device_info udi;
	struct ezdata *firmware;

	while ((ch = getopt(argc, argv, "f:")) != -1) {
		switch (ch) {
		case 'f':
			devname = optarg;
			break;

		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 0)
		usage();

	if (devname == NULL)
		scan_all();

	/* assert(devname != NULL) */

	if ((fd = open(devname, O_WRONLY)) == -1)
		err(EX_OSERR, "%s", devname);

	if (ioctl(fd, USB_GET_DEVICEINFO, &udi) == -1)
		err(EX_OSERR, "%s: Can't get device info", devname);

	if ((firmware = get_firmware(&udi)) == NULL)
		errx(EX_OSERR, "%s: Not a pre-firmware device!", devname);

	flash(fd, firmware);

	return EX_OK;
}

static void
usage(void)
{
	extern char *__progname;

	(void)fprintf(stderr, "usage: %s [-f device]\n", __progname);
	exit(EX_USAGE);
}

static struct ezdata *
get_firmware(struct usb_device_info *udi)
{
	const struct ugen_info *p;

	for (p = tbl; p->firmware != NULL; p++)
		if (p->vendor == udi->udi_vendorNo &&
		    p->product == udi->udi_productNo)
			break;

	return p->firmware;
}

static void
flash(int fd, struct ezdata *firmware)
{
	struct usb_ctl_request ucr;
	u_int8_t rst;

	ucr.ucr_request.bmRequestType = UT_WRITE_VENDOR_DEVICE;
	ucr.ucr_request.bRequest = ANCHOR_LOAD_INTERNAL;
	USETW(ucr.ucr_request.wIndex, 0);
	ucr.ucr_flags = 0;

	rst = ANCHOR_RESET;
	USETW(ucr.ucr_request.wValue, ANCHOR_CPUCS_REG);
	USETW(ucr.ucr_request.wLength, 1);
	ucr.ucr_data = &rst;
	if (ioctl(fd, USB_DO_REQUEST, &ucr) == -1)
		err(EX_OSERR, "Can't enter reset mode");

	(void)usleep(250000);

	for (; firmware->length > 0; firmware++) {
		USETW(ucr.ucr_request.wValue, firmware->address);
		USETW(ucr.ucr_request.wLength, firmware->length);
		ucr.ucr_data = firmware->data;
		if (ioctl(fd, USB_DO_REQUEST, &ucr) == -1)
			err(EX_OSERR, "Can't send firmware chunk");
	}

	(void)usleep(250000);

	rst = 0;
	USETW(ucr.ucr_request.wValue, ANCHOR_CPUCS_REG);
	USETW(ucr.ucr_request.wLength, 1);
	ucr.ucr_data = &rst;
	if (ioctl(fd, USB_DO_REQUEST, &ucr) == -1)
		err(EX_OSERR, "Can't leave reset mode");
}

static void
scan_controller(int ctl)
{
	struct usb_device_info udi;
	struct ezdata *firmware;
	char devname[50];
	int fd;
	int i;

	for (i = 1; i < USB_MAX_DEVICES; i++) {
		udi.udi_addr = i;
		if (ioctl(ctl, USB_DEVICEINFO, &udi) == -1) {
			if (errno != ENXIO)
				warn("Can't get device info (addr %u)", i);
			continue;
		}

		if ((firmware = get_firmware(&udi)) == NULL)
			continue;

		if (strncmp(udi.udi_devnames[0], "ugen", 4) != 0)
			continue;

#if defined(__NetBSD__) || defined(__OpenBSD__)
		(void)snprintf(devname, sizeof devname, "/dev/%s.00",
		    udi.udi_devnames[0]);
#elif defined(__FreeBSD__)
		(void)snprintf(devname, sizeof devname, "/dev/%s",
		    udi.udi_devnames[0]);
#endif
		(void)printf("Pre-firmware modem found at %s!\n", devname);

		if ((fd = open(devname, O_WRONLY)) == -1)
			err(EX_OSERR, "%s", devname);

		flash(fd, firmware);

		exit(EX_OK);
	}
}

static void
scan_all(void)
{
	char buf[50];
	int fd;
	int i;

	for (i = 0; i < 10; i++) {
		(void)snprintf(buf, sizeof buf, "/dev/usb%d", i);

		if ((fd = open(buf, O_RDONLY)) == -1) {
			if (errno != ENOENT && errno != ENXIO)
				warn("%s", buf);
			continue;
		}

		scan_controller(fd);
		(void)close(fd);
	}

	errx(EX_OSERR, "Sorry, no pre-firmware modem found!");
}
